import numpy as np
from sympy import *
from sympy.physics.quantum import TensorProduct
from functools import reduce
from qiskit.quantum_info import *
from scipy.optimize import minimize
from numpy.testing import assert_array_almost_equal as aae
from typing import *

def default_import(items:list = ['a', 'r', 'g', 'm', 'q']):
    
    import_statements = {'a': 'from FrozenYoghourt import *\n', 
                         'r': 'from FrozenYoghourt.mode import *\n', 
                         'g': 'from FrozenYoghourt.gates import *\n', 
                         'm': 'from FrozenYoghourt.maths import *\n', 
                         'q': 'from FrozenYoghourt.quantum import *\n'}
    
    
    import_str = ''
    for item in items:
        import_str += import_statements[item]
        
    print(import_str)
    
def view(mat, rounding = 10):
    if type(mat) == np.ndarray:
        display(Matrix(np.round(mat, rounding)))
    else:
        display(mat.evalf(rounding))
        
   


from FrozenYoghourt import *
from FrozenYoghourt.mode import *


def mm(*lst, no_times=1):
    lst *= no_times
    return reduce(lambda a, b: a @ b, lst)

def tp(*lst, no_times=1):
    lst *= no_times
    if Mode.representation == 'numpy':
        return reduce(lambda a, b: np.kron(a, b), lst)
    else:
        return TensorProduct(*lst)

def svd(A:np.ndarray):
    if Mode.representation == 'numpy':
        return np.linalg.svd(A)
    else:
        AH = A.H
        m, n = A.shape
        if m >= n:
            V, S = (AH * A).diagonalize()

            S_diag = [S[i, i] for i in range(S.rows)]
            ranked = []
            for i, x in enumerate(S.diagonal()):
                if not x.is_zero:
                    ranked.append(i)

            V = V[:, ranked]
            S = Matrix.diag([sqrt(S[i, i]) for i in range(S.rows) if i in ranked])

            V, _ = V.QRdecomposition()
            U = A * V * S.inv()
        else:
            U, S = (A * AH).diagonalize()

            ranked = []
            for i, x in enumerate(S.diagonal()):
                if not x.is_zero:
                    ranked.append(i)

            U = U[:, ranked]
            S = Matrix.diag([sqrt(S[i, i]) for i in range(S.rows) if i in ranked])

            U, _ = U.QRdecomposition()
            V = AH * U * S.inv()

        return U, S, V

def to_su(u:Union[np.ndarray, List[np.ndarray], MutableDenseMatrix, List[MutableDenseMatrix]]):
    if Mode.representation == 'numpy':
        if type(u) == list:
            to_su_list = [mat * complex(np.linalg.det(mat)) ** (-1 / np.shape(mat)[0]) for mat in u]
            return to_su_list
        else:
            return u * complex(np.linalg.det(u)) ** (-1 / np.shape(u)[0])
    else:
        if type(u) == list:
            to_su_list = [mat * complex(mat.det()) ** (-1 / mat.shape[0]) for mat in u]
            return to_su_list
        else:
            return u * complex(u.det()) ** (-1 / u.shape[0])

def fast_substitution(matrix:Matrix, variables, values, to_numpy=False):
    substituted_matrix = matrix.subs(list(zip(variables, values)))
    if to_numpy:
        return np.array(substituted_matrix).astype(np.complex)
    else:
        return substituted_matrix

def optimizing_identities(lhs_fun, rhs_fun, no_var, var_split, random=(0, 2 * np.pi), num_qubits=2, error=1e-7):
    compute_rhs = True
    if var_split == no_var:
        compute_rhs = False

    def cost(x, compute_rhs):
        LHS = lhs_fun(*x[:var_split])

        if compute_rhs:
            RHS = rhs_fun(*x[var_split:])
        else:
            RHS = rhs_fun

        M = (LHS) @ np.conj(RHS).T
        M = M / np.exp(1j * np.angle(M[0, 0]))

        loss = np.linalg.norm(M) - 2 ** (num_qubits / 2)

        return loss

    if random is not None:
        x = np.random.uniform(*random, no_var)
    else:
        x = np.zeros(no_var)

    job = minimize(fun=cost, x0=x, args=(compute_rhs))

    if np.abs(job.fun) < error:
        print("Optimization Sucessful")
    else:
        print("Optimization Unsucessful")

    return job



from FrozenYoghourt import *
    

class Mode:
    representation = 'numpy'

    @classmethod
    def toggle(cls):
        cls.representation = {'numpy': 'sympy',
                              'sympy': 'numpy'}[cls.representation]
        print(f'\033[1m{cls.representation.capitalize()}\033[0m mode activated')

    @classmethod
    def now(cls):
        print(f'\033[1m{cls.representation.capitalize()}\033[0m mode is \033[1mon\033[0m')



from FrozenYoghourt import *
from FrozenYoghourt.mode import *
from FrozenYoghourt.maths import *
from FrozenYoghourt.gates import *


def random_local_ops(num_qubits, no_ops = 1):
    if Mode.representation == 'numpy':
        if no_ops == 1:
            return tp(*[random_unitary(2).data for i in range(num_qubits)])
        else:
            ops_list = [tp(*[random_unitary(2).data for i in range(num_qubits)]) for i in range(no_ops)]
            return ops_list
    else:
        if no_ops == 1:
            return tp(*[U(*symbols(f'theta_{i}, phi_{i}, lambda_{i}')) for i in range(num_qubits)])
        else:
            ops_list = [tp(*[U(*symbols(f'theta_{i}, phi_{i}, lambda_{i}')) 
                                   for i in range(j*num_qubits, (j+1)*num_qubits)]) for j in range(no_ops)]
            return ops_list

def epsilon(psi, num_qubits=2):
    if Mode.representation == 'numpy':
        return np.abs(psi.T @ tp(Y(), no_times=num_qubits) @ psi)[0]
    else:
        return Abs(psi.T @ tp(Y(), no_times=num_qubits) @ psi)[0]

def global_phase(A, B):
    D = np.diag(np.conj(A).T @ B)

    if np.isclose(np.min(D), np.max(D)):
        return D[0]
    else:
        print('A and B are not equivalent up to global phase')
        return False

def kron_decomp(U:np.ndarray):
    m = U.reshape(2, 2, 2, 2).transpose(0, 2, 1, 3).reshape(4, 4)

    u, sv, vh = np.linalg.svd(m)

    a = np.sqrt(sv[0]) * u[:, 0].reshape(2, 2)
    b = np.sqrt(sv[0]) * vh[0, :].reshape(2, 2)

    return a, b


def gamma(U):
    n = int(np.log2(U.shape[0]))

    E = tp(Y(), no_times=n)
    return U @ E @ U.T @ E

def chi(M, variable='x'):
    dim = M.shape[0]

    coef = np.array([1])
    Mk = np.array(M)

    for k in range(1, dim + 1):
        ak = -Mk.trace() / k
        coef = np.append(coef, ak)
        Mk += np.diag(np.repeat(ak, dim))
        Mk = np.dot(M, Mk)

    if Mode.representation == 'numpy':
        return coef

    else:
        var = symbols(variable)
        variable_mat = Matrix([var ** n for n in range(dim + 1)])
        coef = Matrix(coef).T
        char_poly = simplify(coef @ variable_mat)[0]
        return char_poly

def double_cosets(g, h):
    E = Magic()

    g, h = to_su(g), to_su(h)

    u, v = np.conj(E).T @ g @ E, np.conj(E).T @ h @ E

    D1, P1 = np.linalg.eig(u @ u.T)
    D2, P2 = np.linalg.eig(v @ v.T)

    idx = D1.argsort()[::-1]
    D1 = D1[idx]
    P1 = P1[:, idx]

    idx = D2.argsort()[::-1]
    D2 = D2[idx]
    P2 = P2[:, idx]

    if not np.isclose(D1[0], D2[0]):
        D2 = -D2
        idx = D2.argsort()[::-1]
        D2 = D2[idx]
        P2 = P2[:, idx]

    a = P1.T
    b = P2.T

    c = np.conj(v).T @ b.T @ a @ u

    left_coset = E @ b.T @ a @ np.conj(E).T
    right_coset = E @ c.T @ np.conj(E).T

    return left_coset, right_coset





from setuptools import setup, find_packages
 
classifiers = [
  'Development Status :: 5 - Production/Stable',
  'Intended Audience :: Education',
  'Operating System :: Microsoft :: Windows :: Windows 10',
  'License :: OSI Approved :: MIT License',
  'Programming Language :: Python :: 3'
]
 
setup(
  name='FrozenYoghourt',
  version='0.0.11',
  description='Numerical and Symbolic Manipulation for Quantum Computing',
  long_description=open('README.md').read() + '\n\n' + open('CHANGELOG.txt').read(),
  url='',  
  author='Peter Montgomery',
  author_email='petermontgomery056@gmail.com',
  license='MIT', 
  classifiers=classifiers,
  keywords='quantum computing', 
  packages=find_packages(),
  install_requires=[''] 
)